/* create.c */

#include "df3util.h"

static char const * const TU = "create";

static df3Axes aaxes;  /* arg volume axes */
static size_t abpv;    /* arg volume bpv */
static char *atgt;     /* arg target name */


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := axes bpv target.  no change when -input.
 */

static bool procArgs(char *av[]) {
  if (!(av[1]  &&  av[2])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return false;
  }

  if (!getArgCoord(av[0], &aaxes)) {
    EMITEMSG(false, "bad axes value(s)");
    return false;
  }
  else if (!df3_check_axes(aaxes)) {
    EMITEMSG(false, "zero axes value(s)");
    return false;
  }

  uint32_t b = 0;
  if (!getArgValue(av[1], &b)) {
    EMITEMSG(false, "bad bpv value");
    return false;
  }
  else if (!df3_check_bpv((size_t)b)) {
    EMITEMSG(false, "bad bpv value");
    return false;
  }
  abpv = (size_t)b;

  char buf[PATH_MAX];
  if (!getArgFname(av[2], buf)) {
    EMITEMSG(false, "bad target file name");
    return false;
  }
  else if (!mkTargetFname(buf, false, &atgt)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(buf), appLastError(app_lastec));
    return false;
  }

  return true;
}


/* procFile()
 * -------------------------------------------------------------------------- *
 * TODO document.
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Error ec;
  df3Handle * const df3 = vp;

  if (!inpVoxelValid(df3->axes, df3->bpv, at, v))
    return false;
  else if (ecOk != (ec = df3_put_voxel(df3, at, v))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            at.x, at.y, at.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static inline bool procInp(df3Handle *df3) {
  return inputGet(NULL, cback, df3);
}

static bool procFile(void) {
  df3Error ec;
  df3Handle df3 = DF3_INITHANDLE;
  df3.axes = aaxes;
  df3.bpv = abpv;
  df3.fname = atgt;
  if (ecOk != (ec = df3_create_file(&df3))) {
    EMITEMSG(false, "libdf3: %s", df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3, app_logfp);
  }

  bool f = true;
  if (app_fillval)
    f = prepFilled(TU, &df3);

  if (f  &&  itNone != app_inptype) {
    if (app_diagnose)
      EMITDMSG(TU, "reading from <stdin>");
    f = procInp(&df3);
    if (app_diagnose)
      EMITDMSG(TU, "end source <stdin>");
  }

// TODO update app_lastec ??  would clobber procInp()s.
  if (ecOk != df3_close_file(&df3))
    f = false;

  if (app_diagnose)
    EMITDMSG(TU, "target handle closed: %s",
            df3_handle_error(&df3, false));

  return f;
}


/* ------------------------------------------------------------------------- */

static void createTidyup(void) {
  FREENULL(atgt);
}

bool createMain(char *av[]) {
  if (atexit(createTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!isValidFillval(abpv)) {
    return false;
  }
  else if (!haveDiskSpace(aaxes, abpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec = ec_diskFull));
    return false;
  }

  bool const f = procFile();
  if (!f) {
    (void)unlink(atgt);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(atgt));
  }

  if (f  &&  app_verbose)
    printf("created %svolume '%s'.\n",
            app_fillval ? "filled " : "",
            trimmedFn(atgt));

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
