/* import.c */

#include "df3util.h"

static char const * const TU = "import";

static char aaxis;      /* axis arg */
static size_t aoffs;    /* n arg */

static df3Handle df3o;

static char *pngname;   /* source (png) name, alloc'd */


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := axis:n source.
 */

static bool procArgs(char *av[]) {
  if (!(av[0]  &&  av[1])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return false;
  }

  int n = 0;
  if (!getArgAxisN(av[0], &aaxis, &n)) {
    EMITEMSG(false, "bad axis:n argument");
    return false;
  }
  else if (0 > n) {
    EMITEMSG(false, "negative axis:n argument");
    app_lastec = ec_badArgument;
    return false;
  }
  aoffs = (size_t)n;

  char buf[PATH_MAX];
  if (!getArgFname(av[1], buf)) {
    EMITEMSG(false, "bad source file name");
    return false;
  }
  else if (!mkSourceFname(buf, true, false, &pngname)) {
    EMITEMSG(false, "source '%s': %s",
            trimmedFn(buf), appLastError(app_lastec));
    return false;
  }

  return true;
}


/* openTarget()
 * -------------------------------------------------------------------------- *
 * open existing DF3 for writing,
 * bpv must be 1|2, and aoffs must fit given axis.
 */

static bool openTarget(char const * const fn) {
  df3Error ec;

  if (!mkSourceFname(fn, false, true, &df3o.fname)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(fn), appLastError(app_lastec));
    return false;
  }
  else if (ecOk != (ec = df3_open_file(&df3o, false))) {
    EMITEMSG(false, "opening target '%s': %s",
            trimmedFn(df3o.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  bool f = false;
  if (4 == df3o.bpv) {
    EMITEMSG(false, "bad target bpv");
    app_lastec = ec_badArgument;
  }
  else if (badAcol(df3o.axes, aaxis, aoffs)) {
    EMITEMSG(false, "axis '%c' range 0,..,%zu",
            aaxis, axisMax(df3o.axes, aaxis) - 1);
    app_lastec = ec_badArgument;
  }
  else
    f = true;

  if (!f)
    (void)df3_close_file(&df3o);

  return f;
}


/* procFile()
 * -------------------------------------------------------------------------- *
 * allocate + fill buffer, call pngGet(), transfer buffer to target.
 * true if all went well.
 */

static bool writeBuffer(char const axis, size_t const n, void *buf) {
  size_t cols, rows;
  mkBufferXY(df3o.axes, axis, &cols, &rows);
  uint8_t * const p1 = buf;
  uint16_t * const p2 = buf;
  bool const t = 2 == df3o.bpv;
  for (size_t y = 0; rows != y; ++y)
    for (size_t x = 0; cols != x; ++x) {
      size_t const i = y * cols + x;
      uint32_t v = t ? *(p2 + i) : *(p1 + i);
      df3Coord at;
      switch (axis) {
        case 'x': at.x = n; at.y = y; at.z = x; break;
        case 'y': at.x = x; at.y = n; at.z = y; break;
        case 'z': at.x = x; at.y = y; at.z = n; break;
        default:  APPPANIC;
      }
      df3Error ec = df3_put_voxel(&df3o, at, v);
      if (ecOk != ec) {
        EMITEMSG(false, "writing (%zu,%zu) to <%zu,%zu,%zu>: %s",
                x, y, at.x, at.y, at.z, df3_error_message(ec));
        app_lastec = le2ae(ec);
        return false;
      }
    }
  return true;
}

static bool procFile(void) {
  size_t cols, rows;
  mkBufferXY(df3o.axes, aaxis, &cols, &rows);

  size_t const n = cols * rows;
  void *buf = malloc((2 == df3o.bpv ? sizeof(uint16_t) : sizeof(uint8_t)) * n);
  if (!buf) {
    int const e = errno;
    EMITEMSG(false, "create image buffer: %s", strerror(e));
    app_lastec = ec_sysError;
    return false;
  }
  else if (app_diagnose)
    EMITDMSG(TU, "preparing image buffer x:%zu y:%zu @ %zu-bits",
            cols, rows, 8 * df3o.bpv);

  for (size_t i = 0; n != i; ++i)
    if (2 == df3o.bpv)
      *((uint16_t *)buf + i) = (uint16_t)getFillValue(df3o.bpv);
    else
      *((uint8_t *)buf + i) = (uint8_t)getFillValue(df3o.bpv);

  bool f = pngGet(pngname, cols, rows, df3o.bpv, buf);
  if (!f)
    app_lastec = ec_sysError;  /* TODO pngwr needs to do this */

  if (f)
    f = writeBuffer(aaxis, aoffs, buf);

  FREENULL(buf);

  return f;
}


/* ------------------------------------------------------------------------- */

static void importTidyup(void) {
  FREENULL(df3o.fname);
  FREENULL(pngname);
}

bool importMain(char *av[]) {
  if (atexit(importTidyup))
    APPPANIC;

  if (!procArgs(av+1))
    return false;
  else if (!openTarget(av[0]))
    return false;
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  bool f = procFile();
// TODO app_lastec update?
  df3Error const ec = df3_close_file(&df3o);
  if (ecOk != ec) {
    EMITEMSG(false, "closing target '%s': %s",
            trimmedFn(df3o.fname), df3_error_message(ec));
    f = false;
  }

  if (app_diagnose)
    EMITDMSG(TU,"target handle closed: %s",
            df3_handle_error(&df3o, false));

  if (f  &&  app_verbose)
    printf("imported image '%s' into layer %c[%zu].\n",
           trimmedFn(pngname), aaxis, aoffs);

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
