/* main.c */

/* cf setregid(3), srandom(3) */
#define _BSD_SOURCE

#include "df3util.h"

#include <argz.h>
#include <getopt.h>
#include <locale.h>
#include <malloc.h>
#include <sys/resource.h>
#include <time.h>
#include <unistd.h>


static char const * const TU = "main";

/* re help + version */
#define ANRET  __attribute__((noreturn))

/* who we are and cli stuff */
char const app_name[] = "df3util";
char const app_version[] = "0.3.8";

static struct option const app_opts_long[] = {
  {"diagnose",  no_argument,       0, 'D'},
  {"directory", required_argument, 0, 'd'},
  {"fill",      required_argument, 0, 'f'},
  {"gamma",     required_argument, 0, 'g'},
  {"help",      optional_argument, 0, 'h'},
  {"input",     required_argument, 0, 'i'},
  {"logfile",   required_argument, 0, 'l'},
  {"mapfile",   required_argument, 0, 'm'},
  {"seed",      required_argument, 0, 's'},
  {"verbose",   no_argument,       0, 'v'},
  {"version",   no_argument,       0, 'V'},
  {"warnings",  no_argument,       0, 'w'},
  {NULL, 0, 0, CHRNUL}
};

static char const * const app_opts_short = "+d:Df:g:h::i:l:m:s:vVw";

static char const * const app_opts_help[] = {
  "\toutput resource usage and way-point info",
  "path\twhere to write target files",
  "n\t\tvalue given to created voxel data",
  "n.n\tvalue for this screen/display",
  "[=command]\tdisplay this message, or command summary, and exit",
  "fmt\t<stdin> data format: 'csv', 'tcl', or 'txt'",
  "file\twhere to append diagnostic and error messages",
  "file\tcontaining a map for recode",
  "n\t\tstart value for the random number generator",
  "\toutput feedback on completion",
  "\tdisplay the program version and exit",
  "\tdowngrade select errors to warnings",
  NULL
};

typedef bool (*appCmdFn)(char *[]);
typedef struct appCmds appCmds;
struct appCmds {
  char const * const cmd;
  appCmdFn const func;
};

bool copyMain(char *[]);
bool createMain(char *[]);
bool dfdMain(char *[]);
bool exportMain(char *[]);
bool importMain(char *[]);
bool infoMain(char *[]);
bool mergeMain(char *[]);
bool mirrorMain(char *[]);
bool recodeMain(char *[]);
bool rotateMain(char *[]);
bool shiftMain(char *[]);

/* cf info cpp section 3.5 */
#define APPCMDS(name_) { #name_, name_ ## Main }

static appCmds const app_commands[] = {
  APPCMDS(copy),
  APPCMDS(create),
  APPCMDS(dfd),
  APPCMDS(export),
  APPCMDS(import),
  APPCMDS(info),
  APPCMDS(merge),
  APPCMDS(mirror),
  APPCMDS(recode),
  APPCMDS(rotate),
  APPCMDS(shift),
  {NULL, NULL}
};

static appCmdFn app_cmd;

static char const * const app_inpstrings[] = {
  "unknown", "csv", "dfd", "txt", "tcl", NULL
};

static uint32_t app_rngseed;

/* used across TUs */
bool app_diagnose;
long int app_fillval;
double app_gamma;
inpTypes app_inptype;
Ecodes app_lastec;
FILE *app_logfp;
char *app_mapfile;
bool app_verbose;
bool app_warns;
char *app_workdir;


/* commandIndex()
 * -------------------------------------------------------------------------- *
 * returns app_commands[] index, or -1.
 */

static int commandIndex(char const * const s) {
  for (size_t i = 0; app_commands[i].cmd; ++i)
    if (!strcmp(s, app_commands[i].cmd))
      return (int)i;
  return -1;
}


/* note no point setting {app_lastec}, all false returns exit */

static bool storeFillValue(char const * const arg) {
  if ('r' == arg[0]) {
    app_fillval = -1L;  /* (r)andom */
    return true;
  }
  uint32_t v = 0;
  if (!getArgValue(arg, &v)) {
    EMITEMSG(false, "bad fill argument");
    return false;
  }
  app_fillval = (long)v;
  return true;
}


static bool storeInputType(char const * const arg) {
  for (size_t i = 1; app_inpstrings[i]; ++i)
    if (!strcmp(arg, app_inpstrings[i])) {
      app_inptype = (inpTypes)i;
      return true;
    }
  EMITEMSG(false, "bad input type argument");
  return false;
}


static bool storeMapFilename(char const * const arg) {
  char junk[PATH_MAX];
  if (!getArgFname(arg, junk)) {
    EMITEMSG(false, "bad map filename argument");
    return false;
  }
  else if (!statFileExists(arg)) {
    EMITEMSG(false, "no (usable) map file");
    return false;
  }
  if (!(app_mapfile = strdup(arg)))
    APPPANIC;
  return true;
}


static bool storeGammaValue(char const * const arg) {
  char *end = NULL;
  float v;

  errno = 0;
  v = strtof(arg, &end);

  if ((end  &&  *end)  ||  errno) {
    EMITEMSG(false, "bad gamma value");
    return false;
  }
  else if (0.1 > v  ||  9.9 < v) {
    EMITEMSG(false, "gamma value range");
    return false;
  }
  app_gamma = v;
  return true;
}


static bool storeSeedValue(char const * const arg) {
  uint32_t v = 0;
  bool const f = getArgValue(arg, &v);
  if (!(f  &&  v)) {
    EMITEMSG(false, "bad rng seed argument");
    return false;
  }
  app_rngseed = v;
  return true;
}


static bool storeWorkDir(char const * const path) {
  size_t n = strlen(path);
  if (!n  ||  PTHNET <= n) {
    EMITEMSG(false, "bad directory argument");
    return false;
  }
  else if (!statDirectoryPerms(path)) {
    EMITEMSG(false, "no (usable) directory");
    return false;
  }
  else if (PTHSEP == *(path + n - 1))
    n -= 1;

  if (!(app_workdir = strndup(path, n)))
    APPPANIC;

  return true;
}


static bool openLogFile(char const * const fn) {
  FILE *tmp = fopen(fn, "a");
  if (!tmp) {
    int const ec = errno;
    EMITEMSG(false, "error opening log '%s': %s", trimmedFn(fn), strerror(ec));
    return false;
  }
  app_logfp = tmp;
  return true;
}


/* commandHelp(), usageHelp(), usageVersion()
 * -------------------------------------------------------------------------- *
 * usageHelp/Version as mandated by GNU.  commandHelp outputs extended help
 * for the given command.
 */

#include "cmdxhelp.inc"

static void commandHelp(int const) ANRET;
static void commandHelp(int const ix) {
  printf("%s", app_cmd_xhelp[ix]);
  exit(EXIT_SUCCESS);
}

static void usageExit(int const) ANRET;
static void usageExit(int const ec) {
  char const * const fmt1 = "%s - DF3 data file utility.\n"
          "\n  usage: %s [option(s)] command <command-specific>\n"
          "\noptions:\n";

  printf(fmt1, app_name, app_name);

  for (size_t i = 0; app_opts_long[i].name; ++i)
    printf("  -%c | -%s %s.\n",
            app_opts_long[i].val, app_opts_long[i].name, app_opts_help[i]);

  printf("\ncommands (cannot abbreviate):\n  ");

  for (size_t i = 0; app_commands[i].cmd; ++i)
    if (app_commands[i+1].cmd)
      printf("%s, ", app_commands[i].cmd);
    else
      printf("%s.\n", app_commands[i].cmd);

  char const * const fmt2 = "\nsee '%s -help=command' "
                            "and 'man %s' for details.\n";

  printf(fmt2, app_name, app_name);

  exit(ec);
}

static void versionExit(void) ANRET;
static void versionExit(void) {
  char const * const fmt = "%s %s\ncopyright (c) 2017 jr.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to -- whatever.\n"
          "there is no warranty.\n";
  printf(fmt, app_name, app_version);
  exit(EXIT_SUCCESS);
}


/* processCmdline()
 * -------------------------------------------------------------------------- *
 * processes options + command, checks at least one more arg.
 * does not return if error.
 * TODO when store of a given option argument fails returns either 1 or 2,
 *   ought to be one value, but cf man EXIT STATUS.
 */

static void processCmdline(int ac, char *av[], int *ax) {
  bool havedir, havefill, havegamma, haveinp, havelog, havemap, haveseed;

  havedir = havefill = havegamma = haveinp
          = havelog = havemap = haveseed = false;
  opterr = 0;

  while (1) {
    int idx = 0;
    int c = getopt_long_only(ac, av, app_opts_short, app_opts_long, &idx);
    if (-1 == c) break;
    switch (c) {
      case 'D': app_diagnose = true; break;
      case 'd':
        if (havedir) usageExit(2);
        else if (!storeWorkDir(optarg)) exit(1);
        havedir = true;
        break;
      case 'f':
        if (havefill) usageExit(2);
        else if (!storeFillValue(optarg)) exit(2);
        havefill = true;
        break;
      case 'g':
        if (havegamma) usageExit(2);
        else if (!storeGammaValue(optarg)) exit(2);
        havegamma = true;
        break;
      case 'h':
        if (!optarg) usageExit(0);
        else if (-1 == commandIndex(optarg)) usageExit(2);
        commandHelp(commandIndex(optarg));
      case 'i':
        if (haveinp) usageExit(2);
        else if (!storeInputType(optarg)) exit(2);
        haveinp = true;
        break;
      case 'l':
        if (havelog) usageExit(2);
        else if (!openLogFile(optarg)) exit(1);
        havelog = true;
        break;
      case 'm':
        if (havemap) usageExit(2);
        else if (!storeMapFilename(optarg)) exit(1);
        havemap = true;
        break;
      case 's':
        if (haveseed) usageExit(2);
        else if (!storeSeedValue(optarg)) exit(2);
        haveseed = true;
        break;
      case 'V': versionExit();      break;
      case 'v': app_verbose = true; break;
      case 'w': app_warns = true;   break;
      default:  usageExit(2);       break;
    }
  }

  /* need command and at least one more arg */
  if (2 > (ac - optind))
    usageExit(2);
  int const i = commandIndex(av[optind]);
  if (-1 == i)
    usageExit(2);
  app_cmd = app_commands[i].func;

  /* dfd + others do not mix */
  if (haveinp) {
    /* magic number == dfd 3rd in list */
    if (2 == i  &&  itDfd != app_inptype)
      usageExit(2);
    else if (2 != i  &&  itDfd == app_inptype)
      usageExit(2);
  }

  *ax = optind + 1;

  return;
}


/* setupMain(), tidyupMain()
 * -------------------------------------------------------------------------- *
 */

static void tidyupMain(void) {
  FREENULL(app_mapfile);
  FREENULL(app_workdir);
  (void)fclose(app_logfp);
  return;
}

static void setupMain(void) {
  if (!setlocale(LC_ALL, ""))
    APPPANIC;
  else if (setreuid(getuid(), getuid()))
    APPPANIC;
  else if (setregid(getgid(), getgid()))
    APPPANIC;
  else if (atexit(tidyupMain))
    APPPANIC;
  app_logfp = stderr;
  app_inptype = itNone;
  setupGrp();
  setupRE();
  char *tmp = getenv(ENVVAR);
  if (!tmp)
    app_gamma = 2.2;
  else if (!storeGammaValue(tmp)) {
    EMITEMSG(true, "bad environment variable '%s', ignored", ENVVAR);
    app_gamma = 2.2;
  }
  return;
}


static void reportInvoke(char *av[]) {
  char tb[MBUFSZ], td[PTHNET], ts[20];
  char *astring = NULL;
  size_t alen = 0;
  time_t const t = time(NULL);
  struct tm *lt = localtime(&t);
  if (!lt)
    APPPANIC;
  else if (0 == strftime(ts, sizeof(ts), "%d.%m.%Y %H:%M.%S", lt))
    APPPANIC;
  else if (!getcwd(td, PTHNET))
    APPPANIC;

  for (size_t i = 0; av[i]; ++i)
    if (argz_add(&astring, &alen, av[i]))
      APPPANIC;
  argz_stringify(astring, alen, ' ');

  int const n = snprintf(tb, MBUFSZ,
          "system datim: %s\n  cwd: %s\n  args: %s", ts, td, astring);
  if ((MBUFSZ - 1) <= n)
    strcpy(&tb[MBUFSZ-3], "..");  /* forms ellipsis with EOS dot */
  EMITDMSG(TU, "%s", tb);
  free(astring);
  return;
}

static void reportUsage(void) {
  int const w = 20;
  struct mallinfo mi = mallinfo();
  struct rusage ru;
  int junk AUU = getrusage(RUSAGE_SELF, &ru);
// TODO conditionals?  some depend on kernel version.
  EMITDMSG(TU,"resource usage and status.\n"
              "%*s: %ld.%06ld\n"
              "%*s: %ld.%06ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %ld\n"
              "%*s: %d\n"
              "%*s: %d (%s)", 
          w, "user mode time", ru.ru_utime.tv_sec, ru.ru_utime.tv_usec,
          w, "kernel mode time", ru.ru_stime.tv_sec, ru.ru_stime.tv_usec,
          w, "max RSS used", ru.ru_maxrss * 1024L,
          w, "page faults", ru.ru_minflt,
          w, "page faults (IO)", ru.ru_majflt,
          w, "FS in count", ru.ru_inblock,
          w, "FS out count", ru.ru_oublock,
          w, "context switch", ru.ru_nvcsw,
          w, "context switch (inv)", ru.ru_nivcsw,
          w, "bytes allocated", mi.arena,
          w, "last status", app_lastec, appLastError(app_lastec));
  if (stderr != app_logfp)
    fprintf(app_logfp, "\n");
  return;
}


/* ------------------------------------------------------------------------- */

int main(int argc, char *argv[]) {

  setupMain();

  int argx = 0;
  processCmdline(argc, argv, &argx);

  if (app_diagnose)
    reportInvoke(argv+1);

  srandom(app_rngseed ? app_rngseed : (unsigned int)time(NULL));

  bool const f = app_cmd(&argv[argx]);

  if (app_diagnose)
    reportUsage();

  return f ? EXIT_SUCCESS : (ec_badUsage == app_lastec ? 2 : EXIT_FAILURE);
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
