/* regex.c */

#include "df3util.h"

#include <regex.h>

/* offset low limit */
#define LOLIM  -65533

#define REAXIS  "[xyz]"
/* floats cf Levine */
#define REEXP   "([Ee][-+]?[:digit:]]+)"
/* POSIX.1 filename chars */
#define REFNAM  "[A-Za-z[:digit:]_./-]+"
#define RENUM   "[[:digit:]]{1,10}"
#define RESINT  "[+-]?[[:digit:]]{1,5}"
#define REUINT  "[[:digit:]]{1,5}"
#define RESEP   "[,:]"
#define RESPC   "[ \t]"
#define RESR    "[,:-]"
/* dfd.  need min line length of 512 */
#define REDFKK  "[[:alpha:]_]{4,9}"
#define REDFKV  "[[:print:]]{1,499}"
#define REDFCLS "[]]"
#define REDFOPN "[[]"

#define REFLAGS (REG_EXTENDED | REG_NOSUB)

enum reIdx {
  re_axisn = 0,
  re_coord,
  re_fname,
  re_number,
  re_offs,
  re_range,
  re_value,
  re_voxcs,
  re_voxss,
  re_dfd1,
  re_dfd2,
  re_flt1,
  re_flt2,
  re_endRE
};

static char const * const mask[re_endRE] = {
  "%c%*c%d",
  "%zu%*c%zu%*c%zu",
  "%s",
  "%d",
  "%d%*c%d%*c%d",
  "%lu%*c%lu",
  "%lu",
  "%zu%*c%zu%*c%zu%*c%lu",
  "%zu%*c%zu%*c%zu%*c%lu",
  "[%[A-Za-z_]]",
  "[%[A-Za-z_]]%*c%[^\n]",
  "%lf",
  "%lf"
};

static char const * const rexp[re_endRE] = {
  "^" REAXIS "(" RESEP RESINT ")?" "$",
  "^" REUINT RESEP REUINT RESEP REUINT "$",
  "^" REFNAM "$",
  "^" RESINT "$",
  "^" RESINT RESEP RESINT RESEP RESINT "$",
  "^" RENUM RESR RENUM "$",
  "^" RENUM "$",
  "^" REUINT RESEP REUINT RESEP REUINT RESEP RENUM "$",
  "^" REUINT RESPC REUINT RESPC REUINT RESPC RENUM "\n$",
  "^" REDFOPN REDFKK REDFCLS "\n$",
  "^" REDFOPN REDFKK REDFCLS RESPC REDFKV "\n$",
  "^" "[01][.][[:digit:]]*" REEXP "?$",
  "^" "[.][[:digit:]]+" REEXP "?$"
};


/* checkPattern(), setupRE(), tidyupRE().
 * -------------------------------------------------------------------------- *
 */

static regex_t re[re_endRE];

static void tidyupRE(void) {
  for (size_t i = 0; (size_t)re_endRE != i; ++i)
    regfree(&re[i]);
  return;
}

void setupRE(void) {
  for (size_t i = 0; (size_t)re_endRE != i; ++i)
    if (regcomp(&re[i], rexp[i], REFLAGS))
      APPPANIC;
  if (atexit(tidyupRE))
    APPPANIC;
  return;
}

static bool checkPattern(char const * const s, enum reIdx const i) {
  bool f = true;
  if (!strlen(s)  ||  regexec(&re[(size_t)i], s, 0, NULL, 0)) {
    f = false;
    app_lastec = ec_badArgument;
  }
  return f;
}


/* getArgXxxx().
 * -------------------------------------------------------------------------- *
 */

bool getArgAxisN(char const * const a, char *axis, int *n) {
  char tmpa;
  int tmpn = 0;
  if (!checkPattern(a, re_axisn)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (1 == strlen(a))
    tmpa = a[0];
  else if (2 != sscanf(a, mask[(size_t)re_axisn], &tmpa, &tmpn))
    APPPANIC;
  else if (LOLIM > tmpn  ||  UINT16_MAX < tmpn) {
    app_lastec = ec_badArgument;
    return false;
  }
  *axis = tmpa;
  *n = tmpn;
  return true;
}


/* TODO error code bad coord */
bool getArgCoord(char const * const a, df3Coord *v) {
  size_t x, y, z;
  if (!checkPattern(a, re_coord)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (3 != sscanf(a, mask[(size_t)re_coord], &x, &y, &z))
    APPPANIC;
  else if (UINT16_MAX < x  ||  UINT16_MAX < y  ||  UINT16_MAX < z) {
    app_lastec = ec_badAxes;
    return false;
  }
  v->x = x;  v->y = y;  v->z = z;
  return true;
}


bool getArgFname(char const * const a, char *s) {
  if (!checkPattern(a, re_fname)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (1 != sscanf(a, mask[(size_t)re_fname], s))
    APPPANIC;
  return true;
}


bool getArgFloat(char const * const a, float *v) {
  double d = 0.0;
  if (!checkPattern(a, re_flt1))
    if (!checkPattern(a, re_flt2))  {
      app_lastec = ec_badArgument;
      return false;
    }
  /* same mask */
  if (1 != sscanf(a, mask[(size_t)re_flt1], &d))
    APPPANIC;
  else if (0 > d  ||  1 < d) {
    app_lastec = ec_badArgument;
    return false;
  }
  *v = (float)d;
  return true;
}


bool getArgNumber(char const * const a, int *n) {
  int tmp;
  if (!checkPattern(a, re_number)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (1 != sscanf(a, mask[(size_t)re_number], &tmp))
    APPPANIC;
  else if (LOLIM > tmp  ||  UINT16_MAX < tmp) {
    app_lastec = ec_badArgument;
    return false;
  }
  *n = tmp;
  return true;
}


bool getArgOffset(char const * const a, Offset *o) {
  int x, y, z;
  if (!checkPattern(a, re_offs)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (3 != sscanf(a, mask[(size_t)re_offs], &x, &y, &z))
    APPPANIC;
  else if (LOLIM > x  ||  LOLIM > y  ||  LOLIM > z) {
    app_lastec = ec_badOffset;
    return false;
  }
  else if (UINT16_MAX <= x  ||  UINT16_MAX <= y  ||  UINT16_MAX <= z) {
    app_lastec = ec_badOffset;
    return false;
  }
  o->x = x;  o->y = y;  o->z = z;
  return true;
}


/* TODO error code bad range */
bool getArgRange(char const * const a, uint32_t *lo, uint32_t *hi) {
  unsigned long int v1, v2;
  if (!checkPattern(a, re_range)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (2 != sscanf(a, mask[(size_t)re_range], &v1, &v2))
    APPPANIC;
  else if (UINT32_MAX < v1  ||  UINT32_MAX < v2) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (v2 <= v1) {
    app_lastec = ec_badArgument;
    return false;
  }
  *lo = (uint32_t)v1;  *hi = (uint32_t)v2;
  return true;
}


bool getArgValue(char const * const a, uint32_t *v) {
  unsigned long int tmp;
  if (!checkPattern(a, re_value)) {
    app_lastec = ec_badArgument;
    return false;
  }
  else if (1 != sscanf(a, mask[(size_t)re_value], &tmp))
    APPPANIC;
  else if (UINT32_MAX < tmp) {
    app_lastec = ec_badArgument;
    return false;
  }
  *v = (uint32_t)tmp;
  return true;
}


bool getArgVoxel(char const * const a, bool const cs,
                 df3Coord *at, uint32_t *v) {
  enum reIdx const are = cs ? re_voxcs : re_voxss;
  size_t x, y, z;
  unsigned long int val;
  if (!checkPattern(a, are)) {
    app_lastec = ec_badVoxel;
    return false;
  }
  else if (4 != sscanf(a, mask[(size_t)are], &x, &y, &z, &val))
    APPPANIC;
  else if (UINT16_MAX <= x  ||  UINT16_MAX <= y  ||  UINT16_MAX <= z) {
    app_lastec = ec_badOffset;
    return false;
  }
  else if (UINT32_MAX < val) {
    app_lastec = ec_badVoxel;
    return false;
  }
  *at = (df3Coord){x, y, z};
  *v = (uint32_t)val;
  return true;
}


int getEdfdKV(char const * const a, char *k, char *v) {
  if (checkPattern(a, re_dfd1)) {
    if (1 != sscanf(a, mask[(size_t)re_dfd1], k))
      APPPANIC;
    return 1;
  }
  else if (checkPattern(a, re_dfd2)) {
    if (2 != sscanf(a, mask[(size_t)re_dfd2], k, v))
      APPPANIC;
    return 2;
  }
  else
    return 0;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
